# pylint: skip-file
import datetime
import os
import subprocess
import threading
from copy import deepcopy

import botocore
import pytest
from opentaskpy.taskhandlers import transfer
from pytest_shell import fs

from opentaskpy import exceptions
from opentaskpy.addons.aws.remotehandlers.s3 import S3Transfer
from tests.fixtures.localstack import *  # noqa: F403, F405

os.environ["OTF_NO_LOG"] = "0"
os.environ["OTF_LOG_LEVEL"] = "DEBUG"

BUCKET_NAME = "otf-addons-aws-s3-test"
BUCKET_NAME_2 = "otf-addons-aws-s3-test-2"


root_dir_ = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "test",
)


s3_file_watch_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": ".*\\.txt",
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
        "fileWatch": {
            "timeout": 10,
        },
    },
}


s3_age_conditions_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "fileRegex": ".*\\.txt",
        "conditionals": {
            "age": {
                "gt": 10,
                "lt": 18,
            }
        },
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}

s3_file_size_conditions_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "fileRegex": ".*\\.txt",
        "conditionals": {
            "size": {
                "gt": 10,
                "lt": 20,
            }
        },
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}


s3_file_watch_custom_creds_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            "access_key_id": "test",
            "secret_access_key": "test",
            "region_name": "eu-west-1",
        },
        "fileWatch": {
            "timeout": 10,
            "directory": "src",
            "fileRegex": ".*\\.txt",
        },
    },
}


s3_to_s3_copy_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": ".*\\.txt",
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}

s3_to_s3_copy_2_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": "regex-test-5\\.txt",
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}

s3_to_s3_proxy_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": ".*\\.txt",
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "transferType": "proxy",
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}


s3_to_s3_copy_with_fin_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": ".*\\.txt",
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "flags": {
                "fullPath": "dest/my_fin.fin",
            },
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}


s3_to_s3_pca_delete_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": "file-pca-.*\\.txt",
        "postCopyAction": {
            "action": "delete",
        },
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}


s3_to_s3_pca_move_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": "pca-move\\.txt",
        "postCopyAction": {
            "action": "move",
            "destination": "src/archive/",
        },
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}

s3_to_s3_assume_role_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": "file-assumerole\\.txt",
        "postCopyAction": {
            "action": "delete",
        },
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            "assume_role_arn": "arn:aws:iam::01234567890:role/dummy-role",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}


s3_to_s3_pca_rename_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": "file-pca-rename-(.*)\\.txt",
        "postCopyAction": {
            "action": "rename",
            "destination": "src/archive/",
            "pattern": "file-pca-rename-(.*)\\.txt",
            "sub": "file-pca-rename-\\1-renamed.txt",
        },
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}

s3_to_s3_rename_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": "file-rename-(.*)\\.txt",
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "rename": {"pattern": "abc", "sub": "def"},
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}

s3_to_s3_proxy_rename_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": "file-rename-proxy-(.*)\\.txt",
        "protocol": {
            "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME_2,
            "directory": "dest",
            "rename": {"pattern": "abc", "sub": "def"},
            "transferType": "proxy",
            "protocol": {
                "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
            },
        },
    ],
}


s3_to_ssh_copy_task_definition = {
    "type": "transfer",
    "source": {
        "bucket": BUCKET_NAME,
        "directory": "src",
        "fileRegex": ".*\\.txt",
        "protocol": {"name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer"},
    },
    "destination": [
        {
            "hostname": "127.0.0.1",
            "directory": "/tmp/testFiles/dest",
            "protocol": {
                "name": "ssh",
                "port": "3221",
                "credentials": {
                    "username": "application",
                    "keyFile": f"{root_dir_}/testFiles/id_rsa",
                },
            },
        }
    ],
}

ssh_to_s3_copy_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "127.0.0.1",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*\\.txt",
        "protocol": {
            "name": "ssh",
            "port": "3222",
            "credentials": {
                "username": "application",
                "keyFile": f"{root_dir_}/testFiles/id_rsa",
            },
        },
    },
    "destination": [
        {
            "bucket": BUCKET_NAME,
            "directory": "dest",
            "fileRegex": ".*\\.txt",
            "protocol": {"name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer"},
        }
    ],
}


@pytest.fixture(scope="function")
def setup_bucket(credentials, s3_client):
    # This all relies on docker container for the AWS stack being set up and running
    # The AWS CLI should also be installed

    buckets = [BUCKET_NAME, BUCKET_NAME_2]
    # Delete existing buckets and recreate
    for bucket in buckets:
        subprocess.run(
            ["awslocal", "s3", "rb", f"s3://{bucket}", "--force"], check=False
        )
        subprocess.run(["awslocal", "s3", "mb", f"s3://{bucket}"], check=False)


def test_remote_handler():
    transfer_obj = transfer.Transfer(
        None, "s3-file-watch", s3_file_watch_task_definition
    )

    transfer_obj._set_remote_handlers()

    # Validate some things were set as expected
    assert transfer_obj.source_remote_handler.__class__.__name__ == "S3Transfer"

    # dest_remote_handler should be None
    assert transfer_obj.dest_remote_handlers is None


def test_s3_file_watch(s3_client, setup_bucket, tmp_path):
    transfer_obj = transfer.Transfer(
        None, "s3-file-watch", s3_file_watch_task_definition
    )

    # Create a file to watch for with the current date
    datestamp = datetime.datetime.now().strftime("%Y%m%d")

    # Write a test file locally
    fs.create_files(
        [
            {f"{tmp_path}/{datestamp}.txt": {"content": "test1234"}},
            {f"{tmp_path}/{datestamp}1.csv": {"content": "test1234"}},
            {f"{tmp_path}/{datestamp}2.pdf": {"content": "test1234"}},
            {f"{tmp_path}/{datestamp}3.docx": {"content": "test1234"}},
        ]
    )

    with pytest.raises(exceptions.RemoteFileNotFoundError) as cm:
        transfer_obj.run()
    assert "No files found after " in cm.value.args[0]

    # Upload the 3 non-matching files
    for file in os.listdir(tmp_path):
        if not file.endswith(".txt"):
            create_s3_file(s3_client, f"{tmp_path}/{file}", f"src/{file}")

    # This time write the contents after 5 seconds
    t = threading.Timer(
        5,
        create_s3_file,
        [s3_client, f"{tmp_path}/{datestamp}.txt", "src/test.txt"],
    )
    t.start()
    print(  # noqa: T201
        "Started thread - Expect file in 5 seconds, starting task-run now..."
    )

    assert transfer_obj.run()


def test_s3_age_conditions_size(setup_bucket, tmp_path, s3_client):
    transfer_obj = transfer.Transfer(
        None, "s3-age-conditions", s3_age_conditions_task_definition
    )

    # Write a test file locally
    files = [
        {f"{tmp_path}/too_old_file.txt": {"content": "123"}},
        {f"{tmp_path}/correct_file.txt": {"content": "123"}},
        {f"{tmp_path}/too_new_file.txt": {"content": "123"}},
    ]
    fs.create_files(files)
    import time

    for file in files:
        file_name = os.path.basename(list(file.keys())[0])
        create_s3_file(s3_client, list(file.keys())[0], f"src/{file_name}")
        # Sleep 6 seconds
        time.sleep(6)

    assert transfer_obj.run()

    objects = s3_client.list_objects(Bucket=BUCKET_NAME_2)
    # check that the correct_file.txt is in the bucket, and not the other 2
    assert len(objects["Contents"]) == 1
    assert objects["Contents"][0]["Key"] == "dest/correct_file.txt"


def test_s3_file_conditions_size(setup_bucket, tmp_path, s3_client):
    transfer_obj = transfer.Transfer(
        None, "s3-file-size-conditions", s3_file_size_conditions_task_definition
    )

    # Write a test file locally
    files = [
        {f"{tmp_path}/too_small_file.txt": {"content": "1"}},
        {f"{tmp_path}/correct_file.txt": {"content": "test12345678"}},
        {f"{tmp_path}/too_large_file.txt": {"content": "test12345678901234567890"}},
    ]
    fs.create_files(files)
    for file in files:
        file_name = os.path.basename(list(file.keys())[0])
        create_s3_file(s3_client, list(file.keys())[0], f"src/{file_name}")

    assert transfer_obj.run()

    objects = s3_client.list_objects(Bucket=BUCKET_NAME_2)
    # check that the correct_file.txt is in the bucket, and not the other 2
    assert len(objects["Contents"]) == 1
    assert objects["Contents"][0]["Key"] == "dest/correct_file.txt"


def test_s3_to_s3_copy(setup_bucket, s3_client, tmp_path):
    transfer_obj = transfer.Transfer(None, "s3-to-s3", s3_to_s3_copy_task_definition)

    # Create a file to watch for with the current date
    datestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")

    # Write a test file locally

    fs.create_files([{f"{tmp_path}/{datestamp}.txt": {"content": "test1234"}}])
    create_s3_file(s3_client, f"{tmp_path}/{datestamp}.txt", "src/test.txt")

    assert transfer_obj.run()

    # Check that the file is in the destination bucket
    s3_response = s3_client.head_object(
        Bucket=BUCKET_NAME_2,
        Key="dest/test.txt",
    )
    assert s3_response["ResponseMetadata"]["HTTPStatusCode"] == 200


def test_s3_to_s3_copy_2(setup_bucket, s3_client, tmp_path):
    transfer_obj = transfer.Transfer(None, "s3-to-s3", s3_to_s3_copy_2_task_definition)

    # Ensure the source bucket is empty first, and the dest too
    s3_client.delete_object(Bucket=BUCKET_NAME, Key="src/*")
    s3_client.delete_object(Bucket=BUCKET_NAME_2, Key="dest/*")

    # Create 10x files, but only one that matches the regex for the transfer
    for i in range(10):
        fs.create_files([{f"{tmp_path}/regex-test-{i}.txt": {"content": "test1234"}}])
        create_s3_file(
            s3_client, f"{tmp_path}/regex-test-{i}.txt", f"src/regex-test-{i}.txt"
        )

    assert transfer_obj.run()

    # Check that the file is in the destination bucket
    s3_response = s3_client.head_object(
        Bucket=BUCKET_NAME_2,
        Key="dest/regex-test-5.txt",
    )
    assert s3_response["ResponseMetadata"]["HTTPStatusCode"] == 200

    # Ensure that there are no other files in the dest
    objects = s3_client.list_objects(Bucket=BUCKET_NAME_2)
    assert len(objects["Contents"]) == 1


def test_s3_file_matching(setup_bucket, s3_client, tmp_path):
    # Ensure the source bucket is empty first
    s3_client.delete_object(Bucket=BUCKET_NAME, Key="src/*")

    # Create 10x files, but only one that matches the regex for the transfer exactly.
    # Create a copy of each file in another subdir too, and make sure it doesn't match that path too
    for i in range(10):
        fs.create_files([{f"{tmp_path}/regex-test-{i}.txt": {"content": "test1234"}}])
        create_s3_file(
            s3_client, f"{tmp_path}/regex-test-{i}.txt", f"src/regex-test-{i}.txt"
        )
        create_s3_file(
            s3_client,
            f"{tmp_path}/regex-test-{i}.txt",
            f"src/archive/regex-test-{i}.txt",
        )

    # Create a new S3 remotehandler object
    s3_remote_handler = S3Transfer(s3_to_s3_copy_2_task_definition["source"])

    # Check that only 1 file is matched
    assert (
        len(
            s3_remote_handler.list_files(
                directory=s3_remote_handler.spec["directory"],
                file_pattern=s3_remote_handler.spec["fileRegex"],
            )
        )
        == 1
    )


def test_s3_to_s3_proxy(setup_bucket, s3_client, tmp_path):
    transfer_obj = transfer.Transfer(None, "s3-to-s3", s3_to_s3_proxy_task_definition)

    # Create a file to watch for with the current date
    datestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")

    # Write a test file locally

    fs.create_files([{f"{tmp_path}/{datestamp}.txt": {"content": "test1234"}}])
    create_s3_file(s3_client, f"{tmp_path}/{datestamp}.txt", "src/test.txt")

    assert transfer_obj.run()

    # Check that the file is in the destination bucket
    s3_response = s3_client.head_object(
        Bucket=BUCKET_NAME_2,
        Key="dest/test.txt",
    )
    assert s3_response["ResponseMetadata"]["HTTPStatusCode"] == 200


def test_local_to_s3_proxy(setup_bucket, s3_client, tmp_path):

    # Create a file to watch for with the current date
    datestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")

    # Write a test file locally
    fs.create_files([{f"{tmp_path}/{datestamp}.txt": {"content": "test1234"}}])

    # Override the definition to use local source instead
    local_to_s3_task_definition_copy = deepcopy(s3_to_s3_proxy_task_definition)
    local_to_s3_task_definition_copy["source"] = {
        "directory": f"{tmp_path}",
        "fileRegex": f"{datestamp}\\.txt",
        "protocol": {"name": "local"},
    }

    transfer_obj = transfer.Transfer(
        None, "local-to-s3", local_to_s3_task_definition_copy
    )

    assert transfer_obj.run()

    # Check that the file is in the destination bucket
    s3_response = s3_client.head_object(
        Bucket=BUCKET_NAME_2,
        Key=f"dest/{datestamp}.txt",
    )
    assert s3_response["ResponseMetadata"]["HTTPStatusCode"] == 200


def test_s3_to_s3_invalid_source(setup_bucket, s3_client, tmp_path):
    s3_to_s3_copy_task_definition_copy = deepcopy(s3_to_s3_copy_task_definition)

    s3_to_s3_copy_task_definition_copy["source"]["bucket"] = "invalid-bucket"

    transfer_obj = transfer.Transfer(
        None, "s3-to-s3-invalid-source", s3_to_s3_copy_task_definition_copy
    )

    with pytest.raises(botocore.errorfactory.ClientError):
        transfer_obj.run()


def test_s3_to_s3_invalid_destination(credentials, setup_bucket, s3_client, tmp_path):
    s3_to_s3_copy_task_definition_copy = deepcopy(s3_to_s3_copy_task_definition)

    s3_to_s3_copy_task_definition_copy["destination"][0]["bucket"] = "invalid-bucket"

    # Create a file to find
    datestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")

    # Write a test file locally

    fs.create_files([{f"{tmp_path}/{datestamp}.txt": {"content": "test1234"}}])
    create_s3_file(s3_client, f"{tmp_path}/{datestamp}.txt", "src/test.txt")

    transfer_obj = transfer.Transfer(
        None, "s3-to-s3-invalid-destination", s3_to_s3_copy_task_definition_copy
    )

    with pytest.raises(exceptions.RemoteTransferError):
        transfer_obj.run()


def test_s3_to_s3_with_fin_copy(setup_bucket, tmp_path, s3_client):
    transfer_obj = transfer.Transfer(
        None, "s3-to-s3-with-fin", s3_to_s3_copy_with_fin_task_definition
    )

    # Create a file to watch for with the current date
    datestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")

    # Write a test file locally

    fs.create_files([{f"{tmp_path}/{datestamp}.txt": {"content": "test1234"}}])
    create_s3_file(s3_client, f"{tmp_path}/{datestamp}.txt", "src/test.txt")

    assert transfer_obj.run()

    objects = s3_client.list_objects(Bucket=BUCKET_NAME_2)

    assert len(objects["Contents"]) == 2
    # Check that the file and fin file are in the destination bucket
    object_keys = [obj["Key"] for obj in objects["Contents"]]
    assert "dest/test.txt" in object_keys
    assert "dest/my_fin.fin" in object_keys


def test_s3_to_s3_copy_disable_bucket_owner_acl(setup_bucket, s3_client, tmp_path):
    s3_to_s3_copy_task_definition_copy = deepcopy(s3_to_s3_copy_task_definition)
    s3_to_s3_copy_task_definition_copy["destination"][0]["protocol"][
        "disableBucketOwnerControlACL"
    ] = True
    transfer_obj = transfer.Transfer(
        None, "s3-to-s3", s3_to_s3_copy_task_definition_copy
    )

    # Create a file to watch for with the current date
    datestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")

    # Write a test file locally

    fs.create_files([{f"{tmp_path}/{datestamp}.txt": {"content": "test1234"}}])
    create_s3_file(s3_client, f"{tmp_path}/{datestamp}.txt", "src/test.txt")

    assert transfer_obj.run()

    # Can't really test this works with localstack, since there's no actual IAM permissions set on anything
    # Check that the file is in the destination bucket
    objects = s3_client.list_objects(Bucket=BUCKET_NAME_2)
    assert len(objects["Contents"]) == 1
    assert objects["Contents"][0]["Key"] == "dest/test.txt"


def test_s3_to_s3_assume_role(setup_bucket, tmp_path, s3_client):
    transfer_obj = transfer.Transfer(
        None, "s3-to-s3-assume_role", s3_to_s3_assume_role_task_definition
    )

    # Create a file to watch for with the current date
    # Write a test file locally

    fs.create_files([{f"{tmp_path}/file-assumerole.txt": {"content": "test1234"}}])
    create_s3_file(
        s3_client, f"{tmp_path}/file-assumerole.txt", "src/file-assumerole.txt"
    )

    assert transfer_obj.run()

    objects = s3_client.list_objects(Bucket=BUCKET_NAME_2)
    # check that the correct_file.txt is in the bucket, and not the other 2
    assert len(objects["Contents"]) == 1
    assert objects["Contents"][0]["Key"] == "dest/file-assumerole.txt"

    # Check that the file is not in the source bucket
    objects = s3_client.list_objects(Bucket=BUCKET_NAME)
    assert "Contents" not in objects


def test_s3_to_s3_copy_pca_delete(setup_bucket, tmp_path, s3_client):
    transfer_obj = transfer.Transfer(
        None, "s3-to-s3-pca-delete", s3_to_s3_pca_delete_task_definition
    )

    # Create a file to watch for with the current date
    datestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")

    # Write a test file locally

    fs.create_files([{f"{tmp_path}/file-pca-{datestamp}.txt": {"content": "test1234"}}])
    create_s3_file(
        s3_client,
        f"{tmp_path}/file-pca-{datestamp}.txt",
        f"src/file-pca-{datestamp}.txt",
    )

    assert transfer_obj.run()

    objects = s3_client.list_objects(Bucket=BUCKET_NAME_2)
    # check that the correct_file.txt is in the bucket, and not the other 2
    assert len(objects["Contents"]) == 1
    assert objects["Contents"][0]["Key"] == f"dest/file-pca-{datestamp}.txt"

    # Check that the file is not in the source bucket
    objects = s3_client.list_objects(Bucket=BUCKET_NAME)
    assert "Contents" not in objects


def test_s3_to_s3_copy_pca_move(setup_bucket, tmp_path, s3_client):
    transfer_obj = transfer.Transfer(
        None, "s3-to-s3-pca-move", s3_to_s3_pca_move_task_definition
    )

    # Write a test file locally

    fs.create_files([{f"{tmp_path}/pca-move.txt": {"content": "test1234"}}])
    create_s3_file(s3_client, f"{tmp_path}/pca-move.txt", "src/pca-move.txt")

    assert transfer_obj.run()

    # Check that the file is in the destination bucket
    objects = s3_client.list_objects(Bucket=BUCKET_NAME_2)
    assert len(objects["Contents"]) == 1
    assert objects["Contents"][0]["Key"] == "dest/pca-move.txt"

    # Check that the file has been moved to archive
    objects = s3_client.list_objects(Bucket=BUCKET_NAME)
    assert len(objects["Contents"]) == 1
    assert objects["Contents"][0]["Key"] == "src/archive/pca-move.txt"


def test_s3_to_s3_copy_pca_rename(setup_bucket, tmp_path, s3_client):
    transfer_obj = transfer.Transfer(
        None, "s3-to-s3-pca-rename", s3_to_s3_pca_rename_task_definition
    )

    # Write a test file locally
    fs.create_files([{f"{tmp_path}/file-pca-rename-1234.txt": {"content": "test1234"}}])
    create_s3_file(
        s3_client,
        f"{tmp_path}/file-pca-rename-1234.txt",
        "src/file-pca-rename-1234.txt",
    )

    assert transfer_obj.run()

    # Check that the file is in the destination bucket
    objects = s3_client.list_objects(Bucket=BUCKET_NAME_2)
    assert len(objects["Contents"]) == 1
    assert objects["Contents"][0]["Key"] == "dest/file-pca-rename-1234.txt"

    objects = s3_client.list_objects(Bucket=BUCKET_NAME)
    assert len(objects["Contents"]) == 1
    assert (
        objects["Contents"][0]["Key"] == "src/archive/file-pca-rename-1234-renamed.txt"
    )


def test_s3_to_s3_copy_rename(setup_bucket, tmp_path, s3_client):
    transfer_obj = transfer.Transfer(
        None, "s3-to-s3-rename", s3_to_s3_rename_task_definition
    )

    # Write a test file locally
    fs.create_files([{f"{tmp_path}/file-rename-abc.txt": {"content": "test1234"}}])
    create_s3_file(
        s3_client,
        f"{tmp_path}/file-rename-abc.txt",
        "src/file-rename-abc.txt",
    )

    assert transfer_obj.run()

    # Check that the file is in the destination bucket with new name
    objects = s3_client.list_objects(Bucket=BUCKET_NAME_2)
    assert len(objects["Contents"]) == 1
    assert objects["Contents"][0]["Key"] == "dest/file-rename-def.txt"


def test_s3_to_s3_proxy_rename(setup_bucket, s3_client, tmp_path):
    transfer_obj = transfer.Transfer(
        None, "s3-to-s3-proxy-rename", s3_to_s3_proxy_rename_task_definition
    )

    # Write a test file locally
    fs.create_files(
        [{f"{tmp_path}/file-rename-proxy-abc.txt": {"content": "test1234"}}]
    )
    create_s3_file(
        s3_client,
        f"{tmp_path}/file-rename-proxy-abc.txt",
        "src/file-rename-proxy-abc.txt",
    )

    assert transfer_obj.run()

    # Check that the file is in the destination bucket with new name
    objects = s3_client.list_objects(Bucket=BUCKET_NAME_2)
    assert len(objects["Contents"]) == 1
    assert objects["Contents"][0]["Key"] == "dest/file-rename-proxy-def.txt"


def test_s3_file_watch_custom_creds(
    setup_bucket,
    tmp_path,
    s3_client,
):
    transfer_obj = transfer.Transfer(
        None, "s3-file-watch-custom-creds", s3_file_watch_custom_creds_task_definition
    )

    # Create a file to watch for with the current date
    datestamp = datetime.datetime.now().strftime("%Y%m%d")

    # Write a test file locally
    fs.create_files([{f"{tmp_path}/{datestamp}.txt": {"content": "test1234"}}])
    # Write the dummy file to the test S3 bucket
    create_s3_file(s3_client, f"{tmp_path}/{datestamp}.txt", "src/test.txt")

    assert transfer_obj.run()


def create_s3_file(s3_client, local_file, object_key):
    s3_client.put_object(
        Bucket=BUCKET_NAME,
        Key=object_key,
        Body=open(local_file, "rb"),
    )



from opentaskpy.config.schemas import validate_execution_json

valid_protocol_definition = {
    "name": "opentaskpy.addons.aws.remotehandlers.ecsfargate.FargateTaskExecution",
}

valid_execution = {
    "type": "execution",
    "clusterName": "some_cluster_name",
    "taskFamily": "opentaskpy-aws",
    "networkConfiguration": {
        "awsvpcConfiguration": {
            "subnets": [
                "subnet-0e35b35406e3b9fd6",
            ],
            "securityGroups": ["sg-0eaf9802367ef3b41"],
            "assignPublicIp": "ENABLED",
        }
    },
    "containerOverrides": {
        "command": ["echo", "TEST"],
        "environment": [
            {"name": "TEST", "value": "test_env_var"},
            {"name": "TEST2", "value": "test2"},
        ],
    },
    "cloudwatchLogGroupName": "/aws/batch/job",
    "timeout": 10,
    "protocol": valid_protocol_definition,
}


def test_fargate():
    json_data = {
        "type": "execution",
    }
    # Append properties from valid_execution onto json_data
    json_data.update(valid_execution)

    assert validate_execution_json(json_data)

    # Remove protocol
    del json_data["protocol"]
    assert not validate_execution_json(json_data)



"""AWS S3 remote handler."""

import glob
import os
import re
from datetime import datetime
from time import time

import boto3
import opentaskpy.otflogging
from botocore.exceptions import ClientError
from dateutil.tz import tzlocal
from opentaskpy.remotehandlers.remotehandler import (
    RemoteExecutionHandler,
    RemoteTransferHandler,
)

from .creds import set_aws_creds

MAX_OBJECTS_PER_QUERY = 100


class S3Transfer(RemoteTransferHandler):
    """S3 remote transfer handler."""

    TASK_TYPE = "T"

    def __init__(self, spec: dict):
        """Initialise the S3Transfer handler.

        Args:
            spec (dict): The spec for the transfer. This is either the source, or the
            destination spec.
        """
        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )
        self.aws_access_key_id: str | None = None
        self.aws_secret_access_key: str | None = None
        self.region_name: str | None = None
        self.temporary_creds_expiry: datetime | None = None

        super().__init__(spec)

        set_aws_creds(self)

        self.get_session()

        self.get_s3_client()

    def check_credential_expiry(self) -> None:
        """Check the expiry of the temporary credentials."""
        if self.temporary_creds_expiry:
            self.logger.debug(
                f"Temporary creds expire at: {self.temporary_creds_expiry} - Now: {datetime.now(tz=tzlocal())}"
            )
            if self.temporary_creds_expiry < datetime.now(tz=tzlocal()):
                self.logger.info("Renewing temporary credentials")
                self.get_session()
                self.get_s3_client()

    def get_session(self) -> None:
        """Configures a session object to use for API calls."""
        # Use boto3 to setup the required object for the transfer
        kwargs = {
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "region_name": self.region_name,
        }

        self.session = boto3.session.Session(**kwargs)

    def get_s3_client(self) -> None:
        """Get the temporary credentials, or just return the client."""
        # If there's an override for endpoint_url in the environment, then use that
        kwargs2 = {}
        if os.environ.get("AWS_ENDPOINT_URL"):
            kwargs2["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

        self.sts_client = self.session.client("sts", **kwargs2)

        if self.assume_role_arn:
            self.logger.info(f"Assuming role: {self.assume_role_arn}")
            assumed_role_object = self.sts_client.assume_role(
                RoleArn=self.assume_role_arn,
                RoleSessionName=f"OTF{time()}",
                DurationSeconds=900,
            )
            temporary_creds = assumed_role_object["Credentials"]
            # Set the credentials
            self.aws_access_key_id = temporary_creds["AccessKeyId"]
            self.aws_secret_access_key = temporary_creds["SecretAccessKey"]
            self.temporary_creds_expiry = temporary_creds["Expiration"]

            # Set these in the session
            self.session = boto3.session.Session(
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=temporary_creds["SessionToken"],
                region_name=self.region_name,
            )

        self.s3_client = self.session.client("s3", **kwargs2)

    def supports_direct_transfer(self) -> bool:
        """Return True, as you can do bucket to bucket transfers."""
        return True

    def handle_post_copy_action(self, files: list[str]) -> int:
        """Handle the post copy action specified in the config.

        Args:
            files (list[str]): A list of files that need to be handled.

        Returns:
            int: 0 if successful, 1 if not.
        """
        # Check that our creds are valid
        self.check_credential_expiry()

        # Determine the action to take
        # Delete the files
        if self.spec["postCopyAction"]["action"] == "delete":
            self.logger.info(f"Deleting files: {files}")
            response = self.s3_client.delete_objects(
                Bucket=self.spec["bucket"],
                Delete={
                    "Objects": [{"Key": file} for file in files],
                    "Quiet": True,
                },
            )

            # Check response for errors
            if response.get("Errors"):
                self.logger.error(response)
                return 1

            # Verify the files have been deleted
            for file in files:
                try:
                    response = self.s3_client.head_object(
                        Bucket=self.spec["bucket"], Key=file
                    )
                    self.logger.error(response)
                    self.logger.error(f"Failed to delete file: {file}")
                    return 1
                except ClientError as e:
                    # If it's a 404 then its good
                    if e.response["Error"]["Code"] == "404":
                        continue
                    # Otherwise, it's an error
                    self.logger.exception(e)
                    return 1

        # Copy the files to the new location, and then delete the originals
        if (
            self.spec["postCopyAction"]["action"] == "move"
            or self.spec["postCopyAction"]["action"] == "rename"
        ):
            for file in files:
                new_file = (
                    f"{self.spec['postCopyAction']['destination']}{file.split('/')[-1]}"
                )
                if self.spec["postCopyAction"]["action"] == "rename":
                    # Use the pattern and sub values to rename the file correctly
                    new_file = re.sub(
                        self.spec["postCopyAction"]["pattern"],
                        self.spec["postCopyAction"]["sub"],
                        new_file,
                    )

                self.logger.info(f'"Moving" file from {file} to {new_file}')

                self.s3_client.copy_object(
                    Bucket=self.spec["bucket"],
                    CopySource={
                        "Bucket": self.spec["bucket"],
                        "Key": file,
                    },
                    Key=new_file,
                )

                # Check that the copy worked
                try:
                    self.s3_client.head_object(Bucket=self.spec["bucket"], Key=new_file)
                except Exception as e:
                    # Print the exception message
                    self.logger.error(e)
                    self.logger.error(f"Failed to copy file: {file}")
                    return 1

                response = self.s3_client.delete_objects(
                    Bucket=self.spec["bucket"],
                    Delete={
                        "Objects": [{"Key": file}],
                        "Quiet": True,
                    },
                )
                # Check response for errors
                if response.get("Errors"):
                    self.logger.error(response)
                    return 1

                # Check that the delete worked
                try:
                    response = self.s3_client.head_object(
                        Bucket=self.spec["bucket"], Key=file
                    )
                    self.logger.error(response)
                    self.logger.error(f"Failed to delete file: {file}")
                    return 1
                except ClientError as e:
                    # If it's a 404 then its good
                    if e.response["Error"]["Code"] == "404":
                        continue
                    # Otherwise, it's an error
                    self.logger.exception(e)
                    return 1
        return 0

    def list_files(
        self, directory: str | None = None, file_pattern: str | None = None
    ) -> dict:
        """Return list of files that match the source definition.

        Args:
            directory (str, optional): The directory to search in. Defaults to None.
            file_pattern (str, optional): The file pattern to search for. Defaults to
            None.

        Returns:
            dict: A dict of files that match the source definition.
        """
        kwargs = {
            "Bucket": self.spec["bucket"],
            "MaxKeys": MAX_OBJECTS_PER_QUERY,
        }
        if directory:
            kwargs["Prefix"] = directory
        elif "directory" in self.spec and str(self.spec["directory"]):
            kwargs["Prefix"] = str(self.spec["directory"])

        remote_files = {}

        self.logger.info(
            f"Listing files in {self.spec['bucket']} matching"
            f" {file_pattern}{' in' + (directory or '')}"
        )

        try:  # pylint: disable=too-many-nested-blocks
            while True:
                # Check that our creds are valid
                self.check_credential_expiry()
                response = self.s3_client.list_objects_v2(**kwargs)

                if response["KeyCount"]:
                    for object_ in response["Contents"]:
                        key = object_["Key"]
                        # Get the filename from the key
                        filename = key.split("/")[-1]

                        if file_pattern and not re.match(file_pattern, filename):
                            continue

                        # Also check the directory
                        if directory:
                            # Get the directory from the key (using basename)
                            file_directory = os.path.dirname(key)
                            if directory and file_directory != directory:
                                continue

                        if key.startswith("/"):
                            # Make sure that there is no directory in the key
                            # otherwise skip it too as we dont want anything in a subdir
                            # (as directory is not set)
                            continue

                        self.logger.info(f"Found file: {filename}")

                        # Get the size and modified time
                        file_attr = self.s3_client.head_object(
                            Bucket=self.spec["bucket"], Key=key
                        )

                        remote_files[key] = {
                            "size": file_attr["ContentLength"],
                            "modified_time": file_attr["LastModified"].timestamp(),
                        }

                    # This handles the pagination
                    # if the NextContinuationToken doesn't exist, then we'll break out
                    # of the loop
                    try:
                        kwargs["ContinuationToken"] = response["NextContinuationToken"]
                    except KeyError:
                        break
                break
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.logger.error(f"Error listing files: {self.spec['bucket']}")
            self.logger.exception(e)
            raise e

        return remote_files

    def move_files_to_final_location(self, files: list[str]) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    # When S3 is the destination
    def pull_files(self, files: list[str]) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def push_files_from_worker(
        self, local_staging_directory: str, file_list: dict | None = None
    ) -> int:
        """Push files from the worker to the destination server.

        Args:
            local_staging_directory (str): The local staging directory to upload the
            files from.
            file_list (dict, optional): The list of files to transfer. Defaults to None.

        Returns:
            int: 0 if successful, 1 if not.
        """
        # Check that our creds are valid
        self.check_credential_expiry()

        result = 0
        files = glob.glob(f"{local_staging_directory}/*")
        kwargs = {}
        if self.bucket_owner_full_control:
            kwargs["ACL"] = "bucket-owner-full-control"

        for file in files:
            # Strip the directory from the file
            file_name = file.split("/")[-1]
            # Handle any rename that might be specified in the spec
            if "rename" in self.spec:
                rename_regex = self.spec["rename"]["pattern"]
                rename_sub = self.spec["rename"]["sub"]

                file_name = re.sub(rename_regex, rename_sub, file_name)
                self.logger.info(f"Renaming file to {file_name}")
            self.logger.info(
                f"Transferring file: {file} to"
                f" s3://{self.spec['bucket']}/{self.spec['directory']}/{file_name}"
            )
            try:
                self.s3_client.upload_file(
                    file,
                    self.spec["bucket"],
                    f"{self.spec['directory']}/{file_name}",
                    ExtraArgs=kwargs,
                )
            except Exception as e:  # pylint: disable=broad-exception-caught
                self.logger.error(f"Failed to transfer file: {file}")
                self.logger.exception(e)
                result = 1

        return result

    def pull_files_to_worker(
        self, files: list[str], local_staging_directory: str
    ) -> int:
        """Pull files to the worker.

        Download files from AWS S3 to the local staging directory.

        Args:
            files (list): A list of files to download.
            local_staging_directory (str): The local staging directory to download the
            files to.

        Returns:
            int: 0 if successful, 1 if not.
        """
        # Check that our creds are valid
        self.check_credential_expiry()

        result = 0
        for file in files:
            # Strip the directory from the file
            file_name = file.split("/")[-1]
            self.logger.info(f"Downloading file: {file}")
            try:
                self.s3_client.download_file(
                    self.spec["bucket"],
                    file,
                    f"{local_staging_directory}/{file_name}",
                )
            except Exception as e:  # pylint: disable=broad-exception-caught
                self.logger.error(f"Failed to transfer file: {file}")
                self.logger.exception(e)
                result = 1

        return result

    def transfer_files(
        self,
        files: list[str],
        remote_spec: dict,
        dest_remote_handler: RemoteTransferHandler,
    ) -> int:
        """Transfer files from the source S3 bucket to the destination bucket.

        Args:
            files (dict): A dictionary of files to transfer.
            remote_spec (dict): Not used by this handler.
            dest_remote_handler (RemoteTransferHandler): The remote handler
            for the destination bucket

        Returns:
            int: 0 if successful, if not, then 1
            command.
        """
        # Check that our creds are valid
        self.check_credential_expiry()

        # Check the remote handler, if it's another S3Transfer, then it's simple
        # to do an S3 copy via boto

        result = 0
        for file in files:
            # Strip the directory from the file
            file_name = file.split("/")[-1]
            # Handle any rename that might be specified in the spec
            if "rename" in dest_remote_handler.spec:
                rename_regex = dest_remote_handler.spec["rename"]["pattern"]
                rename_sub = dest_remote_handler.spec["rename"]["sub"]

                file_name = re.sub(rename_regex, rename_sub, file_name)
                self.logger.info(f"Renaming file to {file_name}")
            self.logger.info(
                f"Transferring file: {file} from {self.spec['bucket']} to"
                f" {dest_remote_handler.spec['bucket']}/{file_name}"
            )
            try:
                self.s3_client.copy(
                    {
                        "Bucket": self.spec["bucket"],
                        "Key": file,
                    },
                    dest_remote_handler.spec["bucket"],
                    f"{dest_remote_handler.spec['directory']}/{file_name}",
                )
            except Exception as e:  # pylint: disable=broad-exception-caught
                self.logger.error(f"Error transferring file: {file}")
                self.logger.exception(e)
                result = 1

        return result

    def create_flag_files(self) -> int:
        """Create the flag files on the S3 bucket.

        Returns:
            int: 0 if successful, 1 if not.
        """
        # Check that our creds are valid
        self.check_credential_expiry()

        result = 0

        object_key = self.spec["flags"]["fullPath"]
        self.logger.info(f"Creating flag file: {object_key}")
        kwargs = {
            "Bucket": self.spec["bucket"],
            "Key": object_key,
        }
        if self.bucket_owner_full_control:
            kwargs["ACL"] = "bucket-owner-full-control"

        try:
            self.s3_client.put_object(**kwargs)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.logger.error(f"Error creating flag file: {object_key}")
            self.logger.exception(e)
            result = 1

        return result

    def tidy(self) -> None:
        """Tidy up the S3 client."""
        self.s3_client.close()


class S3Execution(RemoteExecutionHandler):
    """S3 remote execution handler.

    This is a strange one, because it's not really an execution. But it's a way to
    create flag files without having to define an entire transfer specification.
    """

    TASK_TYPE = "E"

    def tidy(self) -> None:
        """Tidy up the S3 client."""
        self.s3_client.close()

    def __init__(self, spec: dict):
        """Initialise the S3Execution handler.

        Args:
            spec (dict): The spec for the execution.
        """
        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )
        self.aws_access_key_id: str | None = None
        self.aws_secret_access_key: str | None = None
        self.region_name: str | None = None
        self.temporary_creds_expiry: datetime | None = None

        super().__init__(spec)

        set_aws_creds(self)
        self.get_session()
        self.get_s3_client()

    def check_credential_expiry(self) -> None:
        """Check the expiry of the temporary credentials."""
        if self.temporary_creds_expiry:
            self.logger.debug(
                f"Temporary creds expire at: {self.temporary_creds_expiry} - Now: {datetime.now(tz=tzlocal())}"
            )
            if self.temporary_creds_expiry <= datetime.now(tz=tzlocal()):
                self.logger.info("Renewing temporary credentials")
                self.get_session()
                self.get_s3_client()

    def get_session(self) -> None:
        """Configures a session object to use for API calls."""
        # Use boto3 to setup the required object
        kwargs = {
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "region_name": self.region_name,
        }

        self.session = boto3.session.Session(**kwargs)

    def get_s3_client(self) -> None:
        """Get the temporary credentials, or just return the client."""
        # If there's an override for endpoint_url in the environment, then use that
        kwargs2 = {}
        if os.environ.get("AWS_ENDPOINT_URL"):
            kwargs2["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

        if self.assume_role_arn:
            self.logger.info(f"Assuming role: {self.assume_role_arn}")
            self.sts_client = self.session.client("sts", **kwargs2)
            assumed_role_object = self.sts_client.assume_role(
                RoleArn=self.assume_role_arn,
                RoleSessionName=f"OTF{time()}",
            )
            temporary_creds = assumed_role_object["Credentials"]
            # Set the credentials
            self.aws_access_key_id = temporary_creds["AccessKeyId"]
            self.aws_secret_access_key = temporary_creds["SecretAccessKey"]
            self.temporary_creds_expiry = temporary_creds["Expiration"]

            # Set these in the session
            self.session = boto3.session.Session(
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=temporary_creds["SessionToken"],
                region_name=self.region_name,
            )

        self.s3_client = self.session.client("s3", **kwargs2)

    # This cannot be long running, so kill doesn't really need to do anything
    def kill(self) -> None:
        """Kill the remote process."""

    def execute(self) -> bool:
        """Execute the remote command.

        Returns:
            bool: True if the command was executed successfully, False otherwise
        """
        result = True

        object_key = self.spec["key"]
        bucket = self.spec["bucket"]

        kwargs = {
            "Bucket": bucket,
            "Key": object_key,
        }
        if self.bucket_owner_full_control:
            kwargs["ACL"] = "bucket-owner-full-control"

        # Use the boto client to write a blank file with this name to the bucket
        try:
            self.s3_client.put_object(**kwargs)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.logger.error(f"Failed to create flag file: {object_key}")
            self.logger.exception(e)
            result = False

        return result



# pylint: skip-file
import logging
import os
from copy import deepcopy

import boto3
import botocore.exceptions
import opentaskpy.otflogging
import pytest
from opentaskpy.taskhandlers import execution

from tests.fixtures.localstack import *  # noqa: F403, F405, F401
from tests.fixtures.moto import *  # noqa: F403, F405, F401

os.environ["OTF_LOG_LEVEL"] = "DEBUG"

logger = opentaskpy.otflogging.init_logging(__name__)

logger.setLevel(logging.DEBUG)

root_dir_ = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "test",
)

fargate_task_network_configuration = {
    "awsvpcConfiguration": {
        "subnets": [
            "subnet-0e35b35406e3b9fd6",
            "subnet-0bc9452610a626e11",
            "subnet-019c1f6d89f495308",
        ],
        "securityGroups": ["sg-0eaf9802367ef3b41"],
        "assignPublicIp": "ENABLED",
    }
}

fargate_execution_task_definition = {
    "type": "execution",
    "clusterName": "test_cluster",
    "taskFamily": "opentaskpy-aws",
    "timeout": 60,
    "cloudwatchLogGroupName": "/aws/batch/job",
    "networkConfiguration": fargate_task_network_configuration,
    "containerOverrides": {
        "command": ["echo", "TEST"],
        "environment": [
            {"name": "TEST", "value": "test_env_var"},
            {"name": "TEST2", "value": "test2"},
        ],
    },
    "protocol": {
        "name": "opentaskpy.addons.aws.remotehandlers.ecsfargate.FargateTaskExecution"
    },
}


@pytest.fixture(scope="function")
def credentials_aws_dev():
    if not os.environ.get("GITHUB_ACTIONS"):
        # Look for a .env file in the root of the project
        env_file = os.path.join(root_dir_, "../.env")
        if os.path.isfile(env_file):
            with open(env_file) as f:
                for line in f:
                    if line.startswith("#"):
                        continue
                    key, value = line.strip().split("=")
                    os.environ[key] = value

    if os.environ.get("GITHUB_ACTIONS"):
        if not os.environ.get("ECS_AWS_ACCESS_KEY_ID"):
            print("ERROR: Missing AWS creds")  # noqa: T201
            assert False

        # Read the AWS credentials from the environment
        os.environ["AWS_ACCESS_KEY_ID"] = os.environ["ECS_AWS_ACCESS_KEY_ID"]
        os.environ["AWS_SECRET_ACCESS_KEY"] = os.environ["ECS_AWS_SECRET_ACCESS_KEY"]
        os.environ["AWS_DEFAULT_REGION"] = os.environ["ECS_AWS_DEFAULT_REGION"]
        if os.environ.get("AWS_ENDPOINT_URL"):
            del os.environ["AWS_ENDPOINT_URL"]


def create_ecs_cluster():
    session = boto3.session.Session()

    client = session.client("ecs")
    # Check to see if the ECS cluster exists
    try:
        response = client.describe_clusters(clusters=["test_cluster"])
        if response["clusters"] and response["clusters"][0]["status"] == "ACTIVE":
            return response["clusters"][0]["clusterName"]
    except botocore.exceptions.ClientError:
        pass

    response = client.create_cluster(
        clusterName="test_cluster",
        capacityProviders=["FARGATE"],
        defaultCapacityProviderStrategy=[{"capacityProvider": "FARGATE"}],
    )
    assert response["cluster"]["clusterName"] == "test_cluster"
    return response["cluster"]["clusterName"]


def create_fargate_task():
    session = boto3.session.Session()

    client = session.client("ecs")

    # Check to see if the task definition exists
    try:
        response = client.describe_task_definition(taskDefinition="opentaskpy-aws")
        if (
            response["taskDefinition"]
            and response["taskDefinition"]["status"] == "ACTIVE"
        ):
            return response["taskDefinition"]["family"]
    except botocore.exceptions.ClientError:
        pass

    response = client.register_task_definition(
        family="opentaskpy-aws",
        containerDefinitions=[
            {
                "name": "opentaskpy-aws",
                "image": "opentaskpy-aws",
                "cpu": 100,
                "memory": 100,
                "essential": True,
            }
        ],
    )

    assert response["taskDefinition"]["family"] == "opentaskpy-aws"

    return response["taskDefinition"]["family"]


def test_remote_handler(credentials):
    execution_obj = execution.Execution(
        None, "run-fargate-task", fargate_execution_task_definition
    )

    execution_obj._set_remote_handlers()

    # Validate some things were set as expected
    assert execution_obj.remote_handlers[0].__class__.__name__ == "FargateTaskExecution"


def test_run_fargate_task(credentials_aws_dev):
    create_ecs_cluster()
    create_fargate_task()

    fargate_object = execution.Execution(
        None, "run-fargate-task", fargate_execution_task_definition
    )

    # Execute the task
    assert fargate_object.run()


def test_run_fargate_task_fail(credentials_aws_dev):
    create_ecs_cluster()
    create_fargate_task()

    # Change the definition to force it to fail
    fargate_execution_task_definition_copy = deepcopy(fargate_execution_task_definition)
    fargate_execution_task_definition_copy["containerOverrides"]["command"] = [
        "task-run",
    ]

    execution_obj = execution.Execution(
        None, "call-fargate-task", fargate_execution_task_definition_copy
    )

    # Execute the task
    assert not execution_obj.run()


# @mock_ecs
# def test_run_mocked_fargate_task(cleanup_credentials, credentials_moto):
#     create_ecs_cluster()
#     create_fargate_task()

#     # Manually create a FargateTaskExecution object
#     fargate_task_execution = FargateTaskExecution(fargate_execution_task_definition)

#     # Patch the boto3 client with a mocked one
#     fargate_task_execution.ecs_client = boto3.client("ecs")

#     # Execute the task. This will always timeout, because the mock never changes from
#     # RUNNING status
#     assert not fargate_task_execution.execute()



from opentaskpy.config.schemas import validate_execution_json

valid_protocol_definition = {
    "name": "opentaskpy.addons.aws.remotehandlers.lambda.LambdaExecution",
}

valid_execution = {
    "functionArn": "arn:aws:lambda:eu-west-1:000000000000:function:my-function",
    "invocationType": "Event",
    "protocol": valid_protocol_definition,
}


def test_lambda():
    json_data = {
        "type": "execution",
    }
    # Append properties from valid_execution onto json_data
    json_data.update(valid_execution)

    assert validate_execution_json(json_data)

    # Remove protocol
    del json_data["protocol"]
    assert not validate_execution_json(json_data)



"""AWS Fargate Task remote handler."""

import os
from datetime import datetime
from time import sleep, time

import boto3
import botocore.exceptions
import opentaskpy.otflogging
from dateutil.tz import tzlocal
from opentaskpy.remotehandlers.remotehandler import RemoteExecutionHandler

from .creds import set_aws_creds


class FargateTaskExecution(RemoteExecutionHandler):
    """AWS Fargate Task remote handler."""

    TASK_TYPE = "E"
    fargate_task_id: str

    def tidy(self) -> None:
        """Tidy up the ecs client."""
        self.ecs_client.close()

    def __init__(self, spec: dict):
        """Initialise the FargateTaskExecution handler.

        Args:
            spec (dict): The spec for the execution.
        """
        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )
        self.aws_access_key_id: str | None = None
        self.aws_secret_access_key: str | None = None
        self.region_name: str | None = None
        self.temporary_creds_expiry: datetime | None = None

        super().__init__(spec)

        set_aws_creds(self)
        self.get_session()

        self.get_ecs_client()

    def check_credential_expiry(self) -> None:
        """Check the expiry of the temporary credentials."""
        if self.temporary_creds_expiry:
            self.logger.debug(
                f"Temporary creds expire at: {self.temporary_creds_expiry} - Now: {datetime.now(tz=tzlocal())}"
            )
            if self.temporary_creds_expiry < datetime.now(tz=tzlocal()):
                self.logger.info("Renewing temporary credentials")
                self.get_session()
                self.get_ecs_client()

    def get_session(self) -> None:
        """Configures a session object to use for API calls."""
        # Use boto3 to setup the required object
        kwargs = {
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "region_name": self.region_name,
        }

        self.session = boto3.session.Session(**kwargs)

    def get_ecs_client(self) -> None:
        """Get the temporary credentials, or just return the client."""
        # If there's an override for endpoint_url in the environment, then use that
        kwargs2 = {}
        if os.environ.get("AWS_ENDPOINT_URL"):
            kwargs2["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

        self.sts_client = self.session.client("sts", **kwargs2)

        if self.assume_role_arn:
            self.logger.info(f"Assuming role: {self.assume_role_arn}")
            assumed_role_object = self.sts_client.assume_role(
                RoleArn=self.assume_role_arn,
                RoleSessionName=f"OTF{time()}",
            )
            temporary_creds = assumed_role_object["Credentials"]
            # Set the credentials
            self.aws_access_key_id = temporary_creds["AccessKeyId"]
            self.aws_secret_access_key = temporary_creds["SecretAccessKey"]

            # Set these in the session
            self.session = boto3.session.Session(
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=temporary_creds["SessionToken"],
                region_name=self.region_name,
            )

        self.ecs_client = self.session.client("ecs", **kwargs2)

    def kill(self) -> None:
        """Kill the fargate task function.

        Runs the stop_task function against AWS ECS. This will terminate the task.
        """
        # We need to kill the running execute function
        self.ecs_client.stop_task(
            cluster=self.spec["clusterName"],
            task=self.task_id,
            reason="Task killed by OTF",
        )

        self.tidy()
        self.logger.info("Closed ECS client")

    def execute(self) -> bool:
        """Execute the Fargate task.

        Triggers the fargate task. This will run the task, and then continue to poll the
        state of the task, until it finishes (errors or succeeds). Or, if a timeout is
        set within the task definition, it will timeout after that time.

        If a cloudwatch log group is defined in the task definition, then the logs will
        be collected and returned in the output.

        Returns:
            bool: True if the task succeeded, False if it failed.
        """
        result = True

        task = self.spec["taskFamily"]
        cluster_name = self.spec["clusterName"]
        timeout = self.spec.get("timeout", -1)
        init_timeout = self.spec.get("initTimeout", 60)

        try:
            overrides = {"containerOverrides": [self.spec["containerOverrides"]]}

            overrides["containerOverrides"][0]["name"] = "default"

            run_response = self.ecs_client.run_task(
                cluster=cluster_name,
                taskDefinition=task,
                overrides=overrides,
                networkConfiguration=self.spec["networkConfiguration"],
                launchType="FARGATE",
            )

            # Check that the task has been triggered successfully
            # Check that we got a 200 OK response
            if run_response["ResponseMetadata"]["HTTPStatusCode"] != 200:
                self.logger.error(
                    f"Failed to run fargate task: {task} in cluster {cluster_name}"
                )
                return False

            # Check that the task has been triggered successfully
            if run_response["failures"]:
                self.logger.error(
                    f"Failed to run fargate task: {task} in cluster {cluster_name}"
                )
                return False

            # Get the task id
            self.fargate_task_id = run_response["tasks"][0]["taskArn"].split("/")[-1]

            # Now we loop until the task has finished, or until we've timed out
            # We'll check the status of the task every 5 seconds
            # If the task has a timeout set, then we'll use that as the timeout
            while True:
                self.check_credential_expiry()
                # Get the task status
                self.logger.info("Checking status of task")
                task_status = self.ecs_client.describe_tasks(
                    cluster=cluster_name, tasks=[self.fargate_task_id]
                )

                # Check that we got a 200 OK response
                if task_status["ResponseMetadata"]["HTTPStatusCode"] != 200:
                    self.logger.error(
                        f"Failed to get status of fargate task: {task} in cluster"
                        f" {cluster_name}"
                    )
                    result = False
                    break

                # Check the status of the task
                task_status = task_status["tasks"][0]["lastStatus"]
                self.logger.info(f"Task status: {task_status}")

                # If the task has not started running yet, start decrementing the
                # init_timer. If the init_timer reaches 0, then we can break out of the
                # loop and return False
                if task_status == "PENDING":
                    if init_timeout <= 0:
                        self.logger.error(
                            f"Task: {task} in cluster {cluster_name} timed out while"
                            " initialising."
                        )
                        result = False
                        break

                    init_timeout -= 5

                # If the task has finished, then we can break out of the loop
                if task_status in ["STOPPED", "DEPROVISIONING"]:
                    break

                # If the task has timed out, then we can break out of the loop
                if timeout != -1 and timeout <= 0:
                    self.logger.error(
                        f"Task: {task} in cluster {cluster_name} timed out"
                    )
                    result = False
                    break

                # If we haven't timed out, then we can sleep for 5 seconds
                # If the timeout is > -1 then we'll decrement the timeout
                if timeout != -1:
                    timeout -= 5

                sleep(5)

            # If we've got this far, then the task has finished. We need to check the
            # status to see if it succeeded or failed, and if necessary, also pull the
            # logs from CloudWatch Logs
            task_status = self.ecs_client.describe_tasks(
                cluster=cluster_name, tasks=[self.fargate_task_id]
            )

            if len(task_status["tasks"][0]["containers"]) == 0:
                # If there's no containers, then the task either failed, or we're
                # running in a mocked environment. Here we will fail the task
                self.logger.error(
                    f"Task: {task} in cluster {cluster_name} failed with no containers"
                )
                return False

            container_name = task_status["tasks"][0]["containers"][0]["name"]

            # Do we need to obtain logs?
            if self.spec.get("cloudwatchLogGroupName"):
                # Get the log events
                log_events = self.session.client("logs").get_log_events(
                    logGroupName=self.spec["cloudwatchLogGroupName"],
                    logStreamName=f"{task}/{container_name}/{self.fargate_task_id}",
                )

                # Get the logs
                logs = []
                for event in log_events["events"]:
                    logs.append(event["message"])

                self.logger.info("Fargate Task output:")
                # Log the messages to the logger
                for line in logs:
                    self.logger.info(line)

            # Check the exitCode of the container to get the result
            exit_code = (
                1
                if "exitCode" not in task_status["tasks"][0]["containers"][0]
                else task_status["tasks"][0]["containers"][0]["exitCode"]
            )

            if exit_code != 0:
                reason = (
                    task_status["tasks"][0]["containers"][0]["reason"]
                    if "reason" in task_status["tasks"][0]["containers"][0]
                    else task_status["tasks"][0]["stoppedReason"]
                )

                self.logger.error(
                    f"Task: {task} in cluster {cluster_name} failed. Reason: {reason}"
                )
                result = False

        except botocore.exceptions.ClientError as e:
            self.logger.error(
                f"Failed to run fargate task: {task} in cluster {cluster_name}"
            )
            self.logger.error(e)
            result = False

        return result



# pylint: skip-file
import pytest
from opentaskpy.config.schemas import validate_transfer_json


@pytest.fixture(scope="function")
def valid_protocol_definition():
    return {
        "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
    }


@pytest.fixture(scope="function")
def valid_protocol_definition_using_keys():
    return {
        "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        "access_key_id": "12345",
        "secret_access_key": "12345",
    }


@pytest.fixture(scope="function")
def invalid_protocol_definition_using_keys_no_secret():
    return {
        "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        "access_key_id": "12345",
    }


@pytest.fixture(scope="function")
def valid_protocol_definition_using_assume_role():
    return {
        "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        "assume_role_arn": "12345",
    }


@pytest.fixture(scope="function")
def valid_protocol_definition_using_assume_role_and_keys():
    return {
        "name": "opentaskpy.addons.aws.remotehandlers.s3.S3Transfer",
        "assume_role_arn": "12345",
        "access_key_id": "12345",
        "secret_access_key": "12345",
    }


@pytest.fixture(scope="function")
def valid_transfer(valid_protocol_definition):
    return {
        "bucket": "test-bucket",
        "directory": "/src",
        "fileRegex": ".*\\.txt",
        "protocol": valid_protocol_definition,
    }


@pytest.fixture(scope="function")
def valid_destination(valid_protocol_definition):
    return {
        "bucket": "test-bucket",
        "directory": "/dest",
        "protocol": valid_protocol_definition,
    }


def test_s3_source_protocol(
    valid_transfer,
    valid_protocol_definition_using_assume_role,
    valid_protocol_definition_using_keys,
    invalid_protocol_definition_using_keys_no_secret,
    valid_protocol_definition_using_assume_role_and_keys,
):
    json_data = {
        "type": "transfer",
        "source": valid_transfer,
    }

    json_data["source"]["protocol"] = valid_protocol_definition_using_assume_role
    assert validate_transfer_json(json_data)

    json_data["source"]["protocol"] = valid_protocol_definition_using_keys
    assert validate_transfer_json(json_data)

    json_data["source"][
        "protocol"
    ] = valid_protocol_definition_using_assume_role_and_keys
    assert validate_transfer_json(json_data)

    json_data["source"]["protocol"] = invalid_protocol_definition_using_keys_no_secret
    assert not validate_transfer_json(json_data)


def test_s3_source_basic(valid_transfer):
    json_data = {
        "type": "transfer",
        "source": valid_transfer,
    }

    assert validate_transfer_json(json_data)

    # Remove protocol
    del json_data["source"]["protocol"]
    assert not validate_transfer_json(json_data)


def test_s3_source_file_watch(valid_transfer):
    json_data = {
        "type": "transfer",
        "source": valid_transfer,
    }

    json_data["source"]["fileWatch"] = {
        "timeout": 10,
        "directory": "src",
        "fileRegex": ".*\\.txt",
    }

    assert validate_transfer_json(json_data)

    # Remove fileRegex
    del json_data["source"]["fileWatch"]["fileRegex"]
    assert validate_transfer_json(json_data)

    # Remove directory
    del json_data["source"]["fileWatch"]["directory"]
    assert validate_transfer_json(json_data)

    # Add watchOnly
    json_data["source"]["fileWatch"]["watchOnly"] = True
    assert validate_transfer_json(json_data)

    # Add error
    json_data["source"]["error"] = True
    assert validate_transfer_json(json_data)


def test_s3_destination(valid_transfer, valid_destination):
    json_data = {
        "type": "transfer",
        "source": valid_transfer,
        "destination": [valid_destination],
    }

    assert validate_transfer_json(json_data)

    # Add flags
    json_data["destination"][0]["flags"] = {"fullPath": "flag.txt"}
    assert validate_transfer_json(json_data)

    # Remove protocol
    del json_data["destination"][0]["protocol"]
    assert not validate_transfer_json(json_data)



"""AWS Lambda remote handler."""

import base64
import json
import os
from time import time

import boto3
import botocore.exceptions
import opentaskpy.otflogging
from opentaskpy.exceptions import InvalidConfigError
from opentaskpy.remotehandlers.remotehandler import RemoteExecutionHandler

from .creds import set_aws_creds


class LambdaExecution(RemoteExecutionHandler):
    """AWS Lambda remote handler."""

    TASK_TYPE = "E"

    def tidy(self) -> None:
        """Tidy up the lambda client."""
        self.lambda_client.close()  # type: ignore[has-type]

    def __init__(self, spec: dict):
        """Initialise the LambdaExecution handler.

        Args:
            spec (dict): The spec for the execution.
        """
        # Ensure that function_arn is defined in the spec
        # This is really handled by the schema checks

        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )

        self.aws_access_key_id: str | None = None
        self.aws_secret_access_key: str | None = None
        self.region_name: str | None = None

        super().__init__(spec)

        if "functionArn" not in self.spec:
            raise InvalidConfigError("functionArn not defined in spec")

        set_aws_creds(self)

        self.get_session()

        # If there's an override for endpoint_url in the environment, then use that
        kwargs2 = {}
        if os.environ.get("AWS_ENDPOINT_URL"):
            kwargs2["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

        # Set the client to only try once. This prevents the lambda function being called more than once
        # and issues when working with batch timeouts if a lambda function takes longer than the batch timeout
        # It still means that the timeout of a lambda function is always at least 60 seconds due to the way boto3's HTTP timeout works
        kwargs2["config"] = botocore.client.Config(retries={"max_attempts": 0})

        self.sts_client = self.session.client("sts", **kwargs2)  # type: ignore[has-type]

        if self.assume_role_arn:
            self.logger.info(f"Assuming role: {self.assume_role_arn}")
            assumed_role_object = self.sts_client.assume_role(
                RoleArn=self.assume_role_arn,
                RoleSessionName=f"OTF{time()}",
                DurationSeconds=960,  # Lambda can't run for more than 15 mins
            )
            temporary_creds = assumed_role_object["Credentials"]
            # Set the credentials
            self.aws_access_key_id = temporary_creds["AccessKeyId"]
            self.aws_secret_access_key = temporary_creds["SecretAccessKey"]

            # Set these in the session
            self.session = boto3.session.Session(
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=temporary_creds["SessionToken"],
                region_name=self.region_name,
            )

        self.lambda_client = self.session.client("lambda", **kwargs2)

    def get_session(self) -> None:
        """Configures a session object to use for API calls."""
        # Use boto3 to setup the required object for the transfer
        kwargs = {
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "region_name": self.region_name,
        }

        self.session = boto3.session.Session(**kwargs)

    def kill(self) -> None:
        """Kill the lambda function.

        This doesn't do a huge amount. Lambda functions don't run for a long period of
        time anyway, so this is more a way to clean up the client nicely.
        """
        # We need to kill the running execute function
        self.tidy()
        self.logger.info("Closed Lambda client")

    def execute(self) -> bool:
        """Execute the lambda function.

        Triggers the lambda function. The synchronous invocation type will cause this
        thread to block. This can be an issue with batch jobs and their timeout. Boto3
        will timeout after 60 seconds regardless, so Lambda functions that are called
        should not be long running. Their own timeout should be less than the batch timeout
        if one is used.

        An async call will not check the status of the lambda function, only if there
        are errors with invoking it.
        """
        result = True

        function_arn = self.spec["functionArn"]
        invocation_type = (
            self.spec["invocationType"] if "invocationType" in self.spec else "Event"
        )
        payload = None
        if "payload" in self.spec:
            payload = self.spec["payload"]

        try:
            invoke_response = self.lambda_client.invoke(
                FunctionName=function_arn,
                InvocationType=invocation_type,
                LogType="Tail",
                Payload=json.dumps(payload),
            )

            self.logger.debug(f"Got status code: {invoke_response['StatusCode']}")
            # Print the response
            self.logger.debug(f"Lambda function response: {invoke_response}")

            # Print the response payload
            self.logger.info(
                f"Got the following response from Lambda invocation: {invoke_response['Payload'].read().decode('utf-8')}"
            )

            if (
                invoke_response["StatusCode"] != 200
                and invocation_type == "RequestResponse"
            ) or (invoke_response["StatusCode"] != 202 and invocation_type == "Event"):
                self.logger.error(f"Failed to run lambda function: {function_arn}")
                return False

            if "FunctionError" in invoke_response:
                self.logger.error(
                    f"Lambda function returned an error: {function_arn} - AWS Exception"
                    f" message: {invoke_response['FunctionError']}"
                )
                return False

            # Log the response if there is one
            if "LogResult" in invoke_response:
                # base64 decode the result
                log_result = base64.b64decode(invoke_response["LogResult"]).decode()

                self.logger.debug(f"Lambda function log: {log_result}")

            # Also see if there's any actual result body
            if "Payload" in invoke_response:
                result_payload = invoke_response["Payload"].read()
                self.logger.debug(f"Lambda function payload: {result_payload}")

        except botocore.exceptions.ClientError as e:
            self.logger.error(f"Failed to run lambda function: {function_arn}")
            self.logger.error(e)
            result = False

        return result
